@@ -26,7 +26,7 @@ import sys
 from decimal import Context, Decimal, localcontext
 from fractions import Fraction
 from functools import reduce
-from inspect import isclass, isfunction
+from inspect import isclass
 from uuid import UUID
 
 import attr
@@ -67,6 +67,7 @@ from hypothesis.internal.floats import (
 from hypothesis.internal.reflection import (
     define_function_signature,
     is_typed_named_tuple,
+    nicerepr,
     proxies,
     required_args,
 )
@@ -93,6 +94,7 @@ from hypothesis.searchstrategy.datetime import (
     TimedeltaStrategy,
 )
 from hypothesis.searchstrategy.deferred import DeferredStrategy
+from hypothesis.searchstrategy.functions import FunctionStrategy
 from hypothesis.searchstrategy.lazy import LazyStrategy
 from hypothesis.searchstrategy.misc import (
     BoolStrategy,
@@ -517,32 +519,49 @@ def floats(
             since="2018-10-10",
         )
 
-    if exclude_min:
-        if min_value is None:
-            raise InvalidArgument("Cannot exclude min_value=None")
-        min_value = next_up(min_value, width=width)
-    if exclude_max:
-        if max_value is None:
-            raise InvalidArgument("Cannot exclude max_value=None")
-        max_value = next_down(max_value, width=width)
+    if exclude_min and (min_value is None or min_value == float("inf")):
+        raise InvalidArgument("Cannot exclude min_value=%r" % (min_value,))
+    if exclude_max and (max_value is None or max_value == float("-inf")):
+        raise InvalidArgument("Cannot exclude max_value=%r" % (max_value,))
+
+    if min_value is not None and (
+        exclude_min or (min_arg is not None and min_value < min_arg)
+    ):
+        min_value = next_up(min_value, width)
+        assert min_value > min_arg or min_value == min_arg == 0  # type: ignore
+    if max_value is not None and (
+        exclude_max or (max_arg is not None and max_value > max_arg)
+    ):
+        max_value = next_down(max_value, width)
+        assert max_value < max_arg or max_value == max_arg == 0  # type: ignore
 
-    check_valid_interval(min_value, max_value, "min_value", "max_value")
     if min_value == float(u"-inf"):
         min_value = None
     if max_value == float(u"inf"):
         max_value = None
 
-    if min_value is not None and min_arg is not None and min_value < min_arg:
-        min_value = next_up(min_value, width)
-        assert min_value > min_arg  # type: ignore
-    if max_value is not None and max_arg is not None and max_value > max_arg:
-        max_value = next_down(max_value, width)
-        assert max_value < max_arg  # type: ignore
-    if min_value is not None and max_value is not None and min_value > max_value:
-        raise InvalidArgument(
+    bad_zero_bounds = (
+        min_value == max_value == 0
+        and is_negative(max_value)
+        and not is_negative(min_value)
+    )
+    if (
+        min_value is not None
+        and max_value is not None
+        and (min_value > max_value or bad_zero_bounds)
+    ):
+        # This is a custom alternative to check_valid_interval, because we want
+        # to include the bit-width and exclusion information in the message.
+        msg = (
             "There are no %s-bit floating-point values between min_value=%r "
             "and max_value=%r" % (width, min_arg, max_arg)
         )
+        if exclude_min or exclude_max:
+            msg += ", exclude_min=%r and exclude_max=%r" % (exclude_min, exclude_max)
+        if bad_zero_bounds:
+            note_deprecation(msg, since="2019-03-19")
+        else:
+            raise InvalidArgument(msg)
 
     if allow_infinity is None:
         allow_infinity = bool(min_value is None or max_value is None)
@@ -653,7 +672,7 @@ def sampled_from(elements):
 @overload
 def sampled_from(elements):
     # type: (Type[enum.Enum]) -> SearchStrategy[Any]
-    # `SearchStrategy[Enum]` is unreliable due to mataclass issues.
+    # `SearchStrategy[Enum]` is unreliable due to metaclass issues.
     pass  # pragma: no cover
 
 
@@ -675,6 +694,12 @@ def sampled_from(elements):
     """
     values = check_sample(elements, "sampled_from")
     if not values:
+        note_deprecation(
+            "sampled_from() with nothing to sample is deprecated and will be an "
+            "error in a future version.  It currently returns `st.nothing()`, "
+            "which if unexpected can make parts of a strategy silently vanish.",
+            since="2019-03-12",
+        )
         return nothing()
     if len(values) == 1:
         return just(values[0])
@@ -1005,7 +1030,7 @@ def text(
     elif isinstance(alphabet, SearchStrategy):
         char_strategy = alphabet
     else:
-        char_strategy = sampled_from(list(alphabet))
+        char_strategy = sampled_from(list(alphabet)) if alphabet else nothing()
         if not isinstance(alphabet, abc.Sequence):
             raise InvalidArgument(
                 "alphabet must be an ordered sequence, or tests may be "
@@ -1264,17 +1289,30 @@ def from_type(thing):
     # refactoring it's hard to do without creating circular imports.
     from hypothesis.searchstrategy import types
 
+    def as_strategy(strat_or_callable, thing):
+        # User-provided strategies need some validation, and callables even more
+        # of it.  We do this in three places, hence the helper function
+        if not isinstance(strat_or_callable, SearchStrategy):
+            assert callable(strat_or_callable)  # Validated in register_type_strategy
+            strategy = strat_or_callable(thing)
+        else:
+            strategy = strat_or_callable
+        if not isinstance(strategy, SearchStrategy):
+            raise ResolutionFailed(
+                "Error: %s was registered for %r, but returned non-strategy %r"
+                % (thing, nicerepr(strat_or_callable), strategy)
+            )
+        if strategy.is_empty:
+            raise ResolutionFailed("Error: %r resolved to an empty strategy" % (thing,))
+        return strategy
+
     if typing is not None:  # pragma: no branch
         if not isinstance(thing, type):
-            # At runtime, `typing.NewType` returns an identity function rather
-            # than an actual type, but we can check that for a possible match
-            # and then read the magic attribute to unwrap it.
-            if (
-                hasattr(thing, "__supertype__")
-                and hasattr(typing, "NewType")
-                and isfunction(thing)
-                and getattr(thing, "__module__", 0) == "typing"
-            ):
+            if types.is_a_new_type(thing):
+                # Check if we have an explicitly registered strategy for this thing,
+                # resolve it so, and otherwise resolve as for the base type.
+                if thing in types._global_type_lookup:
+                    return as_strategy(types._global_type_lookup[thing], thing)
                 return from_type(thing.__supertype__)
             # Under Python 3.6, Unions are not instances of `type` - but we
             # still want to resolve them!
@@ -1298,12 +1336,7 @@ def from_type(thing):
     # lookup may be a strategy or a function from type -> strategy; and we
     # convert empty results into an explicit error.
     if thing in types._global_type_lookup:
-        strategy = types._global_type_lookup[thing]
-        if not isinstance(strategy, SearchStrategy):
-            strategy = strategy(thing)  # type: ignore
-        if strategy.is_empty:  # type: ignore
-            raise ResolutionFailed("Error: %r resolved to an empty strategy" % (thing,))
-        return strategy  # type: ignore
+        return as_strategy(types._global_type_lookup[thing], thing)
     # If there's no explicitly registered strategy, maybe a subtype of thing
     # is registered - if so, we can resolve it to the subclass strategy.
     # We'll start by checking if thing is from from the typing module,
@@ -1317,19 +1350,13 @@ def from_type(thing):
     # type.  For example, `Number -> integers() | floats()`, but bools() is
     # not included because bool is a subclass of int as well as Number.
     strategies = [
-        v if isinstance(v, SearchStrategy) else v(thing)  # type: ignore
+        as_strategy(v, thing)
         for k, v in types._global_type_lookup.items()
         if isinstance(k, type)
         and issubclass(k, thing)
         and sum(types.try_issubclass(k, typ) for typ in types._global_type_lookup) == 1
     ]
-    empty = ", ".join(repr(s) for s in strategies if s.is_empty)
-    if empty:
-        raise ResolutionFailed(
-            "Could not resolve %s to a strategy; consider using "
-            "register_type_strategy" % empty
-        )
-    elif strategies:
+    if strategies:
         return one_of(strategies)
     # If we don't have a strategy registered for this type or any subtype, we
     # may be able to fall back on type annotations.
@@ -1596,7 +1623,7 @@ def decimals(
         special.append(Decimal("Infinity"))
     if allow_infinity or (allow_infinity is min_value is None):
         special.append(Decimal("-Infinity"))
-    return strat | sampled_from(special)
+    return strat | (sampled_from(special) if special else nothing())
 
 
 def recursive(
@@ -2048,11 +2075,8 @@ class DataStrategy(SearchStrategy):
 
     def __not_a_first_class_strategy(self, name):
         raise InvalidArgument(
-            (
-                "Cannot call %s on a DataStrategy. You should probably be "
-                "using @composite for whatever it is you're trying to do."
-            )
-            % (name,)
+            "Cannot call %s on a DataStrategy. You should probably be using "
+            "@composite for whatever it is you're trying to do." % (name,)
         )
 
 
@@ -2163,3 +2187,33 @@ def emails():
     return builds(u"{}@{}".format, local_part, domains()).filter(
         lambda addr: len(addr) <= 254
     )
+
+
+# Mypy can't yet handle default values with generic types or typevars, but the
+# @overload workaround from https://github.com/python/mypy/issues/3737 doesn't
+# work with @composite functions - Mypy can't see that the function implements
+# `(Any, Callable[..., T], SearchStrategy[T]) -> Callable[..., T]`
+
+
+@defines_strategy
+def functions(like=lambda: None, returns=none()):
+    """A strategy for functions, which can be used in callbacks.
+
+    The generated functions will mimic the interface of ``like``, which must
+    be a callable (including a class, method, or function).  The return value
+    for the function is drawn from the ``returns`` argument, which must be a
+    strategy.
+
+    Note that the generated functions do not validate their arguments, and
+    may return a different value if called again with the same arguments.
+
+    Generated functions can only be called within the scope of the ``@given``
+    which created them.  This strategy does not support ``.example()``.
+    """
+    check_type(SearchStrategy, returns)
+    if not callable(like):
+        raise InvalidArgument(
+            "The first argument to functions() must be a callable to imitate, "
+            "but got non-callable like=%r" % (nicerepr(like),)
+        )
+    return FunctionStrategy(like, returns)
